#include <stdlib.h>
#include <stdio.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include "aswm.h"
#include "event_handlers/create.h"
#include "event_handlers/configure.h"
#include "event_handlers/map.h"
#include "event_handlers/unmap.h"
#include "event_handlers/reparent.h"
#include "event_handlers/destroy.h"
#include "event_handlers/property.h"
#include "event_handlers/message.h"
#include "log/error.h"
#include "aswm/log/log.h"

Aswm _aswm;

void aswm_open(const char* zlog_conf) {
	atexit(aswm_close);

	set_up_logger(zlog_conf);

	XSetErrorHandler(aswm_error_handler);

	_aswm.display = XOpenDisplay(NULL);
	_aswm.desktops = malloc(0);
	_aswm.desktops_count = 0;

	if(_aswm.display == NULL) {
		aswm_error("Cannot open X Display. Check the DISPLAY environement variable, and ensure an X server is running for the corresponding DISPLAY.");
		exit(1);
	}
	Window x_root_window = DefaultRootWindow(_aswm.display);
	XWindowAttributes x_root_attributes;
	XGetWindowAttributes(_aswm.display, x_root_window, &x_root_attributes);

	_aswm.root_window = XCreateSimpleWindow(
			_aswm.display, x_root_window,
			0, 0,
			x_root_attributes.width,
			x_root_attributes.height,
			0, 0, 0);

	// User windows are created from the X root window, while ASWM specific
	// windows are created on the ASWM root window
	XSelectInput(_aswm.display, _aswm.root_window, FocusChangeMask | EnterWindowMask | LeaveWindowMask | ButtonPressMask | ButtonReleaseMask);

	XSelectInput(_aswm.display, x_root_window, SubstructureRedirectMask | SubstructureNotifyMask | FocusChangeMask | EnterWindowMask | LeaveWindowMask | ButtonPressMask | ButtonReleaseMask);

	XMapWindow(_aswm.display, _aswm.root_window);
	XSync(_aswm.display, 0);

	aswm_debug("PointerRoot %lu", PointerRoot);
	aswm_debug("X Root window %lu", x_root_window);
	aswm_debug("ASWM Root window %lu", _aswm.root_window);

	aswm_log_tree(aswm_debug, &_aswm, DefaultRootWindow(_aswm.display), 1);
}

void aswm_create_workspace() {
	_aswm.desktops = realloc(_aswm.desktops, (_aswm.desktops_count+1) * sizeof(Desktop));

	XWindowAttributes root_attributes;
	XGetWindowAttributes(_aswm.display, _aswm.root_window, &root_attributes);

	_aswm.desktops[_aswm.desktops_count].root_window = XCreateSimpleWindow(
			_aswm.display, _aswm.root_window,
			0, 0,
			root_attributes.width, root_attributes.height,
			0, 0, 0);
	_aswm.desktops[_aswm.desktops_count].tile_root = XCreateSimpleWindow(
			_aswm.display, _aswm.desktops[_aswm.desktops_count].root_window,
			0, 0,
			root_attributes.width, root_attributes.height,
			0, 0, 0);
	_aswm.desktops[_aswm.desktops_count].stack_root = XCreateSimpleWindow(
			_aswm.display, _aswm.desktops[_aswm.desktops_count].root_window,
			0, 0,
			root_attributes.width, root_attributes.height,
			0, 0, 0);

	_aswm.current_desktop = _aswm.desktops[_aswm.desktops_count];
	_aswm.desktops_count++;
	XSelectInput(_aswm.display, _aswm.current_desktop.tile_root, SubstructureNotifyMask);
	XSelectInput(_aswm.display, _aswm.current_desktop.stack_root, SubstructureNotifyMask);
	XMapWindow(_aswm.display, _aswm.current_desktop.root_window);
	XMapWindow(_aswm.display, _aswm.current_desktop.tile_root);
	aswm_notice("Create workspace %lu", _aswm.current_desktop.root_window);
}

void aswm_event_loop() {
	for (;;) {
		XEvent e;
		XNextEvent(_aswm.display, &e);

		switch(e.type) {
			case CreateNotify:
				// Windows with override_redirect set to True can be safely
				// ignored, according to the X server specification.
				if(e.xcreatewindow.override_redirect == False)
					OnCreateNotify(&_aswm, &e.xcreatewindow);
				else
					aswm_notice("Ignored create notify %lu", e.xcreatewindow.window);
				break;
			case ConfigureRequest:
				OnConfigureRequest(&e.xconfigurerequest);
				break;
			case MapRequest:
				OnMapRequest(&_aswm, &e.xmaprequest);
				break;
			case UnmapNotify:
				OnUnmapNotify(&_aswm, &e.xunmap);
				break;
			case DestroyNotify:
				OnDestroyNotify(&e.xdestroywindow);
				break;
			case ReparentNotify:
				OnReparentNotify(&e.xreparent);
				break;
			case FocusIn:
				aswm_debug("FocusIn %lu", (&e.xfocus)->window);
				break;
			case FocusOut:
				aswm_debug("FocusOut %lu", e.xfocus.window);
				break;
			case EnterNotify:
				int revert_to;
				Window window;
				XGetInputFocus(e.xcrossing.display, &window, &revert_to);
				aswm_debug("Enter %lu", e.xcrossing.window);
				aswm_debug("Input focus %lu", window);
				break;
			case LeaveNotify:
				aswm_debug("Leave %lu", e.xcrossing.window);
				break;
			case ButtonPress:
				aswm_debug("Press %lu", e.xbutton.window);
				break;
			case ButtonRelease:
				aswm_debug("Release %lu", e.xbutton.window);
				break;
			case ClientMessage:
				OnClientMessage(&_aswm, &e.xclient);
				break;
			case PropertyNotify:
				OnPropertyNotify(&_aswm, &e.xproperty);
				break;
			default:
				aswm_debug("Unknown event: %i", e.type);
				break;
					/* LOG(WARNING) << "Ignored event"; */
		}
	}
}

void aswm_close() {
	free(_aswm.desktops);
	tear_down_logger();
	if(_aswm.display != NULL)
		XCloseDisplay(_aswm.display);
}

Window aswm_currently_active() {
	Window active;
	int status;
	XGetInputFocus(_aswm.display, &active, &status);
	return active;
}

void aswm_activate_window(Window window) {
	/* XEvent event; */
	/* event.xclient.display = _aswm.display; */
	/* event.xclient.window = window; */
	/* event.xclient.message_type = XInternAtom(_aswm.display, "_NET_ACTIVE_WINDOW", True); */
	/* event.xclient.format = 32; */
	/* event.xclient.data.l[0] = 1; // Source indication */
	/* event.xclient.data.l[1] = 1; // Timestamp // Currently not set */
	/* event.xclient.data.l[2] = 0; // Requestor's currently active window: TODO */
	
	aswm_notice("Activate window %lu (currently active: %lu)", window, aswm_currently_active());
	/* XSendEvent(_aswm.display, _aswm.root_window, True, 0 [> N. A. <], &event); */
	XSetInputFocus(_aswm.display, window, RevertToParent, CurrentTime);
}
