#include <stdlib.h>
#include <stdio.h>
#include "aswm/mapper/mapper.h"
#include "aswm/log/log.h"
#include <X11/Xutil.h>

bool aswm_is_resizeable(XSizeHints* size) {
	return size->min_width != size->max_width
		|| size->min_height != size->max_height;
}

/**
 * Returns the free tile built in the split process.
 */
void h_split(Display* display, Window current_node, Window window_to_map);

void aswm_tile_window(Display* display, Window current_node, Window window_to_map) {
	h_split(display, current_node, window_to_map);
}

void h_split(Display* display, Window current_tile, Window window_to_map) {
	// Current:
	// +--------------+
	// | current tile |
	// | +--------+   |
	// | | child  |   |
	// | +--------+   |
	// +--------------+
	// Target:
	// +---------------+
	// | current tile  |
	// | +-----------+ |
	// | | child     | |
	// | +-----------+ |
	// | +-----------+ |
	// | | free tile | |
	// | +-----------+ |
	// +---------------+

	XWindowAttributes current_tile_attributes;
	XGetWindowAttributes(display, current_tile, &current_tile_attributes);

	XReparentWindow(display, window_to_map, current_tile, 0, 0);

	Window _root_window; // unused
	Window _parent_window; // unused
	Window* children;
	unsigned int children_count;

	XQueryTree(display, current_tile, &_root_window, &_parent_window, &children, &children_count);

	for(auto i = 0; i < children_count; i++) {
		XMoveResizeWindow(display, children[i],
				0, i * current_tile_attributes.height / children_count,
				current_tile_attributes.width,
				current_tile_attributes.height / children_count);
	}
	XFree(children);
	XMapWindow(display, window_to_map);
}

void resize_children_to_parent(Display* display, Window parent) {
	Window _root_window; // unused
	Window _grand_parent; // unused
	Window* children;
	unsigned int children_count;
	XQueryTree(
			display, parent,
			&_root_window, &_grand_parent,
			&children, &children_count);
	aswm_debug("  Resize children of %lu:\n", parent);
	for(auto i = 0; i < children_count; i++)
		aswm_debug("\t%lu", children[i]);

	// Resize new children of the parent tile
	XWindowAttributes parent_attributes;
	XGetWindowAttributes(display, parent, &parent_attributes);

	for(auto i = 0; i < children_count; i++) {
		// TODO: currently only hsplit supported
		XMoveResizeWindow(display, children[i],
				0,
				i * parent_attributes.height / children_count,
				parent_attributes.width,
				parent_attributes.height / children_count);
	}
	XFree(children);
}

void aswm_untile_window(Display* display, Window parent, Window window) {
	// +-------------------------------+
	// | root                          |
	// |+-------------+ +-------------+|
	// || parent      | |             ||
	// || +--------+  | |             ||
	// || | win 1  |  | |             ||
	// || +--------+  | |             ||
	// || +--------+  | |             ||
	// || | del win|  | |    win 3    ||
	// || +--------+  | |             ||
	// || +--------+  | |             ||
	// || | win 2  |  | |             ||
	// || +--------+  | |             ||
	// |+-------------+ +-------------+|
	// +-------------------------------+

	Window _root_window; // unused
	Window _grand_parent; // unused
	Window* children;
	unsigned int children_count;
	XQueryTree(
			display, parent,
			&_root_window, &_grand_parent,
			&children, &children_count);
	aswm_debug("  Current children of %lu:", parent);
	for(auto i = 0; i < children_count; i++)
		aswm_debug("\t%lu", children[i]);
	for(auto i = 0; i < children_count; i++) {
		if(children[i] == window)
			// If the unmap request does not come from a destroy window, we
			// still want to remove the window from the ASWM hierarchy. It is
			// the responsability of the client to decide to remap it latter or
			// not.
			XReparentWindow(display, window, XDefaultRootWindow(display), 0, 0);
	}

	resize_children_to_parent(display, parent);
}

/**
 * Currently, resize to the max width and max height of any children, but should
 * be reconsidered as it can create black square artifacts if the max width and
 * height are not for the same window.
 */
void resize_stack(Display* display, Window stack) {
	Window root; // unused
	Window parent;
	Window* children;
	unsigned int children_count;
	XQueryTree(display, stack, &root, &parent, &children, &children_count);

	unsigned int width = 0;
	unsigned int height = 0;
	for(auto i = 0; i < children_count; i++) {
		XWindowAttributes child_attributes;
		XGetWindowAttributes(display, children[i], &child_attributes);
		if(child_attributes.map_state != IsUnmapped) {
			if(child_attributes.width > width)
				width = child_attributes.width;
			if(child_attributes.height > height)
				height = child_attributes.height;
		}
	}

	// The current method is always called when there is at least a child, so
	// children is never NULL
	XFree(children);

	XWindowAttributes parent_attributes;
	XGetWindowAttributes(display, parent, &parent_attributes);
	XMoveResizeWindow(display, stack,
			(parent_attributes.width - width) / 2,
			(parent_attributes.height - height) / 2,
			width, height);
}

void aswm_stack_window(Display* display, Window stack, Window window_to_map, int width, int height) {
	XWindowAttributes stack_attributes;
	XGetWindowAttributes(display, stack, &stack_attributes);

	XSetWindowAttributes stacked_attributes;
	stack_attributes.width = width;
	stack_attributes.height = height;
	stack_attributes.win_gravity = CenterGravity;
	XChangeWindowAttributes(display, window_to_map,
			CWWidth | CWHeight | CWWinGravity,
			&stacked_attributes);
	XReparentWindow(display, window_to_map, stack,
			(stack_attributes.width - width) / 2,
			(stack_attributes.height - height) / 2
			);
	// Map before resize so window_to_map can be taken into account in resizing
	// of stack
	XMapWindow(display, window_to_map);

	resize_stack(display, stack);

	if(stack_attributes.map_state == IsUnmapped) {
		XMapWindow(display, stack);
	}

	aswm_activate_window(window_to_map);
}

void aswm_unstack_window(Display* display, Window stack) {
	Window root; // unused
	Window parent;
	Window* children;
	unsigned int children_count;
	XQueryTree(display, stack, &root, &parent, &children, &children_count);

	bool visible_child = false;
	int i = 0;
	while(!visible_child && i < children_count) {
		XWindowAttributes child_attributes;
		XGetWindowAttributes(display, children[i], &child_attributes);
		visible_child = (child_attributes.map_state == IsViewable);
		i++;
	}
	XFree(children);

	if(visible_child) {
		aswm_debug("Resize stack %lu\n", stack);
		resize_stack(display, stack);
	} else {
		aswm_debug("Nothing in stack, unmap %lu\n", stack);
		XWindowAttributes parent_size;
		XGetWindowAttributes(display, parent, &parent_size);

		XMoveResizeWindow(display, stack, 0, 0, parent_size.width, parent_size.height);
		XUnmapWindow(display, stack);
	}
}

bool is_in_stack(Display* display, Window stack_root, Window window) {
	Window _root; // unused
	Window parent;
	Window* _children; // unused
	unsigned int _children_count; // unused

	XQueryTree(display, window, &_root, &parent, &_children, &_children_count);
	return parent == stack_root;
}

void aswm_handle_normal_hints(Display* display, Window stack_root, Window window, XSizeHints* size, long supplied) {
	aswm_info("Handle normal hints for %lu", window);
	XWindowAttributes win_attributes;
	XGetWindowAttributes(display, window, &win_attributes);

	if(win_attributes.map_state != IsUnmapped) {
		// TODO: handle multiple desktops
		if((supplied & (PMinSize | PMaxSize))) {
			if(is_in_stack(display, stack_root, window)) {
				if(aswm_is_resizeable(size)) {
					// Remap to tile root
					XUnmapWindow(display, window);
					XMapWindow(display, window);
				} else {
					XResizeWindow(display, window, size->max_width, size->max_height);
					resize_stack(display, stack_root);
				}
			} else {
				if(!aswm_is_resizeable(size)) {
					// Resize window to its fixed size before stacking it
					XResizeWindow(display, window, size->max_width, size->max_height);

					// XMapWindow does not trigger a MapRequest for the current
					// client, so we need to call aswm_stack_window directly.
					// aswm_stack_window automatically triggers the appropriate
					// unmap from tile when reparenting from tile to stack
					aswm_stack_window(display, stack_root, window, size->max_width, size->max_height);
				}
			}
		}
	}
}

